@@ -273,9 +273,7 @@ elementTy :: forall k v a f proxy. (Vector v a, Index (Peano k) (Dim v), Functor
           => proxy k -> (a -> f a) -> (v a -> f (v a))
 {-# INLINE elementTy #-}
 elementTy _ f v
-  = fmap vector
-  $ inspect (C.cvec v) 
-    (C.lensF (proxy# @(Peano k)) f construct)
+  = inspect v (C.lensF (proxy# @(Peano k)) f construct)
 
 -- | Left fold over vector
 foldl :: Vector v a => (b -> a -> b) -> b -> v a -> b
@@ -607,7 +605,7 @@ izipWithM f v u = fmap vector
 -- | Zip two vector elementwise using monadic function and discard
 --   result
 izipWithM_
-  :: (Vector v a, Vector v b, Vector v c, Applicative f, Vector v (f c))
+  :: (Vector v a, Vector v b, Vector v c, Applicative f)
   => (Int -> a -> b -> f c) -> v a -> v b -> f ()
 {-# INLINE izipWithM_ #-}
 izipWithM_ f xs ys = C.izipWithM_ f (C.cvec xs) (C.cvec ys)
